(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

header {* Matching *}

theory Matching
imports
  Term_Impl
  Option_Monad
begin

definition matcher :: "('f, 'w) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v, 'w) gsubst set" where
  "matcher t l = {\<sigma>. t = l \<cdot> \<sigma>}"

definition matcherL :: "(('f, 'w) term \<times> ('f, 'v) term) set \<Rightarrow> ('f, 'v, 'w) gsubst set" where
  "matcherL ps = {\<sigma>. (\<forall>p \<in> ps. fst p = snd p \<cdot> \<sigma>)}"

abbreviation matcherLi :: "('v \<times> ('f, 'w) term) set \<Rightarrow> ('f, 'v, 'w) gsubst set" where
  "matcherLi ps \<equiv> matcherL ((\<lambda>p. (snd p, Var (fst p))) ` ps)"


text {*A sound and complete matching algorithm.*}

function
  match_term_list :: "(('f, 'w) term \<times> ('f, 'v) term) list \<Rightarrow> ('v \<times> ('f, 'w) term) list option"
where
  "match_term_list [] = Some []"
| "match_term_list ((Var _, Fun _ _) # ps) = None"
| "match_term_list ((Fun f ts, Fun g ls) # ps) = do {
    guard (f = g);
    E \<leftarrow> zip_option ts ls;
    match_term_list (E @ ps)
  }"
| "match_term_list ((t, Var x) # ps) = do {
    E \<leftarrow> match_term_list ps;
    Some ((x, t) # E)
  }"
  by pat_completeness auto

termination 
proof -
  let ?M = "measures [\<lambda>l. list_size (size \<circ> snd) l, length]"
  show ?thesis 
  proof
    fix f::'f and ts :: "('f, 'w) term list" and g and ls :: "('f, 'v) term list" and ps res
    assume "f = g" and dec: "zip_option ts ls = Some res"
    from dec have len: "length ts = length ls" by auto
    from dec have res: "res = zip ts ls" by simp
    from len and res have main: "list_size (size \<circ> snd) (res @ ps) < Suc (Suc (list_size size ls + list_size (size \<circ> snd) ps))"
      by (auto simp: list_size_map[symmetric] simp del: list_size_map)
    show "(res @ ps, (Fun f ts, Fun g ls) # ps) \<in> ?M"
      by (rule measures_less, auto simp: main simp del: list_size_append)
  qed auto
qed

lemma matcherL_Un [simp]:
  "matcherL (ps \<union> qs) = matcherL ps \<inter> matcherL qs"
  by (unfold matcherL_def, blast)

lemma matcherL_insert:
  "matcherL (insert p ps) = {\<sigma>. fst p = snd p \<cdot> \<sigma>} \<inter> matcherL ps"
  by (unfold matcherL_def, blast)

lemma matcherL_zip:
  "length ts = length ls \<Longrightarrow>
   {\<sigma>. \<forall>i<length ts. ts ! i = ls ! i \<cdot> \<sigma>} = matcherL (set (zip ts ls))" 
proof (induct ts ls rule: list_induct2) 
  case Nil show ?case by (auto simp:matcherL_def)
next
  case (Cons t ts l ls) 
  have "{\<sigma>. \<forall>i<length (t # ts). (t # ts) ! i = (l # ls) ! i \<cdot> \<sigma>} = { \<sigma>. t = l \<cdot> \<sigma>} \<inter> {\<sigma>. \<forall>i<length ts. ts ! i = ls ! i \<cdot> \<sigma>}" (is "?l = ?r")
  proof (intro set_eqI iffI)
    fix s
    assume "s \<in> ?l"
    then show "s  \<in> ?r" by force
  next
    fix s 
    assume Prop2: "s \<in> ?r"
    then have one: "t = l \<cdot> s" by simp
    from Prop2 have two: "\<forall>i< length ts. ts ! i = ls ! i \<cdot> s" by simp
    have "\<forall> i < length (t # ts). (t # ts) ! i = (l # ls) ! i \<cdot> s"
    proof (intro allI impI)
      fix i
      assume "i < length (t # ts)"
      with one two show "(t # ts) ! i = (l # ls) ! i \<cdot> s"
        by (cases i, auto)
    qed
    then show "s \<in> ?l" by simp
  qed
  also have "\<dots> = { \<sigma>. t = l \<cdot> \<sigma>} \<inter> matcherL (set (zip ts ls))" using Cons by simp
  finally show ?case by (simp add: matcherL_insert)
qed

lemma matcherL_Fun [simp]:
  "matcherL (insert (Fun f ts, Fun g ls) ps) =
    (if f = g \<and> length ts = length ls then matcherL (set (zip ts ls)) \<inter> matcherL ps
    else {})" 
proof (cases "f = g \<and> length ts = length ls")
  assume "\<not> (f = g \<and> length ts = length ls)"
  then have "f \<noteq> g \<or> length ts \<noteq> length ls" by blast
  then show ?thesis
  proof
    assume "f \<noteq> g" then show ?thesis by (unfold matcherL_def, simp)
  next
    assume "length ts \<noteq> length ls" then show ?thesis by (unfold matcherL_def, auto)
  qed
next
  assume P: "f = g \<and> length ts = length ls"
  then have "{\<sigma>. ts = map (\<lambda> l. l \<cdot> \<sigma>) ls} = {\<sigma>. \<forall>i<length ts. ts ! i = (map (\<lambda> l. l \<cdot> \<sigma>) ls) ! i}"  using list_eq_iff_nth_eq[where xs = ts] by force
  also have "\<dots> = {\<sigma>. \<forall>i<length ts. ts ! i = ls ! i \<cdot> \<sigma>}" using P by auto 
  also have "\<dots> = matcherL (set (zip ts ls))"
    using P matcherL_zip by blast
  finally show ?thesis using P by (simp add: matcherL_insert)
qed

lemma match_term_list_complete:
  "match_term_list ps = None \<Longrightarrow> matcherL (set ps) = {}"
proof (induct rule:match_term_list.induct)
  case (4 t x ps) then show ?case by (cases "match_term_list ps", unfold matcherL_def, auto)
next
  case (3 f ts g ls ps) then show ?case 
  proof (cases "f = g")
    case False then show ?thesis by (unfold matcherL_def, auto)
  next
    case True then show ?thesis
    proof (cases "zip_option ts ls")
      case (Some E)
      then have len: "length ts = length ls" by auto
      from Some have zip: "E = zip ts ls" by auto
      with Some `f = g` 3(1)[of E] 3(2)
        have "matcherL (set (zip ts ls @ ps)) = {}" by simp
      with len show ?thesis by simp
    next
      case None
      then have difflen: "length ts \<noteq> length ls" by simp
      then show ?thesis by simp
    qed
  qed
qed (auto simp: matcherL_def)

lemma match_term_list_sound:
  "match_term_list ps = Some qs \<Longrightarrow> matcherL (set ps) = matcherLi (set qs)" 
proof (induct arbitrary:qs rule:match_term_list.induct)
  case (3 f ts g ls ps) then show ?case 
  proof (cases "f = g")
    case True then show ?thesis
    proof (cases "zip_option ts ls")
      case None with 3 and `f=g` show ?thesis by simp
    next
      case (Some E)
      then have len: "length ts = length ls" by auto
      from Some have zip: "E = zip ts ls" by auto
      with 3 Some True zip have "matcherL (set (zip ts ls @ ps)) = matcherLi (set qs)" by simp
      with len True show ?thesis by simp
    qed
  qed auto
next
  case (4 t x ps) then show ?case 
    by (cases "match_term_list ps", auto simp: matcherL_insert)
qed (auto simp: matcherL_def)

abbreviation rvars :: "(('f, 'w)term \<times> ('f,'v)term) list \<Rightarrow> 'v set" where
  "rvars ps \<equiv> \<Union> ((vars_term \<circ> snd) ` (set ps))"

abbreviation lvars :: "('v \<times> ('f, 'w) term) list \<Rightarrow> 'v set" where
  "lvars ps \<equiv> { x | t x. (x,t) \<in> set ps}"

lemma match_term_list_vars:
  "match_term_list ps = Some qs \<Longrightarrow> rvars ps = lvars qs"
proof (induct arbitrary:qs rule:match_term_list.induct)
  case (4 t x ps) then show ?case by (cases "match_term_list ps", auto)
next
  case (3 f ts g ls ps) then show ?case 
  proof (cases "f = g")
    case False with 3 show ?thesis by auto
  next
    case True then show ?thesis
    proof (cases "zip_option ts ls")
      case None with 3 and `f=g` show ?thesis by simp
    next
      case (Some E)
      then have len: "length ts = length ls" by auto
      from Some have zip: "E = zip ts ls" by auto
      with 3 Some True zip have ind: "rvars (E @ ps) = lvars qs" by auto
      from len have "rvars (zip ts ls) = rvars [(Fun f ts, Fun g ls)]" 
      proof (induct ls arbitrary: ts, simp)
        case (Cons l ls)
        then show ?case by (cases ts, auto)
      qed
      with ind zip show "rvars ((Fun f ts, Fun g ls) # ps) = lvars qs" by auto
    qed
  qed
qed auto

fun
  match_vars :: "('v \<times> ('f, 'w) term) list \<Rightarrow> ('f, 'v, 'w) gsubstL option"
where
  "match_vars [] = Some []"
| "match_vars ((x, t) # ps) =
    (case map_of ps x of 
      None \<Rightarrow>
      (case match_vars ps of
        Some \<sigma> \<Rightarrow> Some ((x, t) # \<sigma>)
      | None \<Rightarrow> None)
    | Some s \<Rightarrow> if t = s then match_vars ps else None)"

lemma match_vars_sound:
  assumes "match_vars ps = Some \<sigma>"
  shows "mk_subst d \<sigma> \<in> matcherLi (set ps)"
proof -
  {
    assume "match_vars ps = Some \<sigma>"
    then have "mk_subst d \<sigma> \<in> matcherLi (set ps)" 
    proof (induct ps arbitrary: \<sigma>)
      case (Cons xt qs)
      obtain x t where Pair: "xt = (x,t)" by force
      show ?case 
      proof (cases "map_of qs x")
        case (Some s)
        show ?thesis 
        proof (cases "s = t")
          case False
          with Some Pair Cons show ?thesis by simp
        next
          case True
          from Some have "(x,s) \<in> set qs" by (rule map_of_SomeD)
          with True Some Pair have "set (Cons xt qs) = set qs" by auto
          with Pair Cons True Some show ?thesis by auto
        qed
      next
        case None
        with Cons Pair show ?thesis 	    
        proof (cases "match_vars qs", simp)
          case (Some \<delta>)
            from Cons Pair None Some
              have id: "mk_subst d \<sigma> = mk_subst d ((x, t)#\<delta>)" by auto
            with Cons Some have ind: "mk_subst d \<delta> \<in> matcherLi (set qs)" by auto
            from None have "\<not> (\<exists> t. (x,t) \<in> set qs)" by (force simp: map_of_eq_None_iff)
            with ind Pair show ?thesis
              unfolding matcherL_def id 
              by (auto simp add: mk_subst_def)
        qed
      qed
    qed (auto simp: matcherL_def)
  }
  with assms show ?thesis by auto
qed

lemma match_vars_complete:
  assumes "\<delta> \<in> matcherLi (set ps)" 
  shows "\<exists>\<sigma>. match_vars ps = Some \<sigma> \<and> (\<forall>x \<in> lvars ps. \<delta> x = mk_subst d \<sigma> x)" 
proof -
  let ?prop = "\<lambda>\<sigma> ps. match_vars ps = Some \<sigma> \<and> (\<forall> x \<in> lvars ps. \<delta> x = mk_subst d \<sigma> x)"
  from assms show "\<exists>\<sigma>. ?prop \<sigma> ps"
  proof (induct ps)
    case (Cons xt qs)
    obtain x t where Pair: "xt = (x,t)" by force
    with Cons have delt: "\<delta> x = t" unfolding matcherL_def by auto
    from Cons have "\<delta> \<in> matcherLi (set qs)" unfolding matcherL_def by auto 
    with Cons obtain \<sigma> where ind: "?prop \<sigma> qs" by force
    show ?case 
    proof (cases "map_of qs x")
      case (Some s)
      then have  "(x,s) \<in> set qs" by (rule map_of_SomeD)
      with Cons have dels: "\<delta> x = s" unfolding matcherL_def by auto
      with ind `(x,s) \<in> set qs` have "\<delta> x = mk_subst d \<sigma> x" by auto
      with dels delt ind Pair Some have "?prop \<sigma> (xt # qs)"  by auto
      then show ?thesis by blast
    next
      case None
      then have "\<not> (\<exists> s. (x,s) \<in> set qs)" by (force simp: map_of_eq_None_iff)
      have "?prop ((x, t)#\<sigma>) (xt # qs)" using ind Pair None delt by (auto simp: mk_subst_def)
      then show ?thesis by blast
    qed
  qed auto
qed

fun
  match_list :: "(('f, 'w) term \<times> ('f, 'v) term) list \<Rightarrow> ('v \<times> ('f, 'w) term) list option"
where
  "match_list tls =
    (case match_term_list tls of 
      None \<Rightarrow> None 
    | Some ps \<Rightarrow> match_vars ps)"

fun match :: "('f, 'w) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v, 'w) gsubst option" where
  "match t l = (case match_list [(t, l)] of
    None \<Rightarrow> None
  | Some bs \<Rightarrow> Some (mk_subst (\<lambda>x. t) bs))"

lemma match_list_sound:
  assumes match: "match_list tls = Some \<sigma>"
  shows "\<And>t l. (t, l) \<in> set tls \<Longrightarrow> l \<cdot> mk_subst d \<sigma> = t"
proof -
  fix t l
  assume tl: "(t,l) \<in> set tls"
  show "l \<cdot> mk_subst d \<sigma> = t"
    using match
  proof (cases "match_term_list tls")
    case (Some ps) note oSome = this
    then show ?thesis using match 
    proof (cases "match_vars ps")
      case (Some \<delta>)
      with match_vars_sound have "mk_subst d \<delta> \<in> matcherLi (set ps)" by auto
      with match_term_list_sound oSome have "mk_subst d \<delta> \<in> matcherL (set tls)" by force
      with tl oSome Some match show ?thesis unfolding matcherL_def by auto
    qed auto
  qed auto
qed

lemma match_sound:
  assumes "match t l = Some \<sigma>" shows "l \<cdot> \<sigma> = t"
proof -
  from assms obtain bs where "match_list [(t, l)] = Some bs"
    and "mk_subst (\<lambda>x. t) bs = \<sigma>" by (auto split: option.splits)
  from match_list_sound [OF this(1), of t l] and this(2)
    show ?thesis by force
qed

lemma match_list_complete:
  assumes "\<And> t l. (t, l) \<in> set tls \<Longrightarrow> l \<cdot> \<delta> = t" 
  shows "\<exists> \<sigma> . match_list tls = Some \<sigma> \<and> (\<forall> x \<in> rvars tls. \<delta> x = mk_subst d \<sigma> x)"
proof -
  from assms have match: "\<delta> \<in> matcherL (set tls)" unfolding matcherL_def by auto
  show ?thesis
  proof (cases "match_term_list tls")
    case None
    then have "matcherL (set tls) = {}" by (rule match_term_list_complete)
    with match show ?thesis by simp
  next
    case (Some ps)
    then have "matcherL (set tls) = matcherLi (set ps)" by (rule match_term_list_sound)
    with match have "\<delta> \<in> matcherLi (set ps)" by auto
    then have "\<exists> \<sigma>. match_vars ps = Some \<sigma> \<and>  (\<forall>x\<in>lvars ps. \<delta> x = mk_subst d \<sigma> x)"
      by (rule match_vars_complete)
    from this obtain \<sigma> where one: "match_vars ps = Some \<sigma> \<and>  (\<forall>x\<in>lvars ps. \<delta> x = mk_subst d \<sigma> x)" ..
    from Some have "rvars tls = lvars ps" by (rule match_term_list_vars)
    with Some one have "match_list tls = Some \<sigma> \<and> (\<forall>x \<in> rvars tls. \<delta> x = mk_subst d \<sigma> x)" by auto
    then show ?thesis by force
  qed
qed

lemma match_complete:
  assumes "l \<cdot> \<delta> = t" 
  shows "\<exists>\<sigma>. match t l = Some \<sigma> \<and> (\<forall> x \<in> vars_term l. \<delta> x = \<sigma> x)"
proof -
  from assms have precond: "\<And> t' l'. (t',l') \<in> set [(t, l)] \<Longrightarrow> l' \<cdot> \<delta> = t'" by simp
  from match_list_complete[OF precond]
  obtain \<sigma> where "match_list [(t, l)] = Some \<sigma>"
    and "\<forall> x \<in> rvars [(t, l)]. \<delta> x = mk_subst (\<lambda>x. t) \<sigma> x" by blast
  then show ?thesis by auto
qed
      
declare match.simps [simp del]

definition matches :: "('f, 'w) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "matches t l = (case match t l of None \<Rightarrow> False | Some _ \<Rightarrow> True)"

lemma matches [simp]:
  "matches t l = (\<exists>\<sigma> . t = l \<cdot> \<sigma>)"
proof (cases "match t l")
  case (Some \<sigma>)
  with match_sound[OF Some] show ?thesis unfolding matches_def by auto
next
  case None
  have "\<not> (\<exists> \<sigma>. t = l \<cdot> \<sigma>)"
  proof (rule notI)
    assume "\<exists> \<sigma> . t = l \<cdot> \<sigma>"
    then obtain \<sigma> where t: "l \<cdot> \<sigma> = t" by auto
    from match_complete[OF t] None show False by auto
  qed
  with None show ?thesis unfolding matches_def by auto
qed

end

